import errno
import logging
from multiprocessing import reduction
import os
import socket

from .base import BaseTransport
from .. import constants


logger = logging.getLogger(__name__)


try:
    rebuild_socket = reduction.rebuild_socket
    reduce_socket = reduction.reduce_socket
except AttributeError:
    rebuild_socket = reduction._rebuild_socket
    reduce_socket = reduction._reduce_socket


class UNIXSocketTransport(BaseTransport):
    """
    Usage:
    
    transport = UNIXSocketTransport()
    
    process, stdout, stderr = transport.run_cmd('ls -al')
    """
    
    def __init__(self, socket_path='/tmp/errand-boy', listen_backlog=5, **kwargs):
        super(UNIXSocketTransport, self).__init__(**kwargs)
        
        self.socket_path = socket_path
        self.listen_backlog = listen_backlog
    
    def connection_to_string(self, connection):
        if isinstance(connection, tuple):
            return str((connection[0].getsockname(), connection[1],))
        return connection.getsockname()
    
    def server_get_connection(self):
        serversocket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        
        try:
            os.remove(self.socket_path)
        except OSError as e:
            logger.exception(e)
        
        serversocket.bind(self.socket_path)
        serversocket.listen(self.listen_backlog)
        
        return serversocket
    
    def server_recv(self, connection, length):
        clientsocket, address = connection
        
        return clientsocket.recv(length)
    
    def server_send(self, connection, data):
        clientsocket, address = connection
        
        clientsocket.sendall(data)
    
    def server_close(self, connection):
        clientsocket, address = connection
        try:
            clientsocket.close()
        except:
            pass
    
    def server_accept(self, connection):
        return connection.accept()
    
    def server_deserialize_connection(self, connection):
        return rebuild_socket(*connection[0][1]), connection[1]
    
    def server_serialize_connection(self, connection):
        return reduce_socket(connection[0]), connection[1]
    
    def client_get_connection(self):
        clientsocket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        
        clientsocket.connect(self.socket_path)
        
        return clientsocket
    
    def client_send(self, connection, data):
        connection.sendall(data)
    
    def client_recv(self, connection, length):
        return connection.recv(length)
    
    def client_close(self, connection):
        try:
            connection.close()
        except:
            pass



import eventlet

from .base import BaseTransport
from .. import constants


class MockTransport(BaseTransport):
    """
    Usage:
    
    transport = MockTransport()
    
    process, stdout, stderr = transport.run_cmd('ls -al')
    """
    def __init__(self, **kwargs):
        super(MockTransport, self).__init__(**kwargs)
    
    def server_get_connection(self):
        return None
    
    def server_recv(self, connection, l):
        in_conn, out_conn = connection
        
        try:
            return in_conn.pop(0)
        except IndexError:
            return b''
    
    def server_send(self, connection, data):
        in_conn, out_conn = connection
        out_conn.append(data)
        eventlet.sleep(0)
    
    def server_accept(self, connection):
        self.connection = list(), list()
        return self.connection
    
    def run_server(self, max_accepts=1000):
        return
    
    def client_get_connection(self):
        return self.connection
    
    def client_send(self, connection, data):
        in_conn, out_conn = connection
        in_conn.append(data)
        eventlet.sleep(0)
    
    def client_recv(self, connection, l):
        in_conn, out_conn = connection
        
        try:
            return out_conn.pop(0)
        except IndexError:
            return b''
    
    def get_session(self):
        connection = self.server_accept(None)
        
        greenthread = eventlet.spawn(self.server_handle_client, connection)
        
        return super(MockTransport, self).get_session()
        



import six
import subprocess

import errand_boy
from errand_boy.transports import base, unixsocket

from .base import mock, BaseTestCase
from .data import commands, get_command_data, get_req, get_resp


class UNIXSocketTransportClientTestCase(BaseTestCase):
    def test_run_cmd(self):
        transport = unixsocket.UNIXSocketTransport()
        
        with self.socket_patcher as socket:
            clientsocket = mock.Mock()
            
            cmd = 'ls -al'
            
            stdout, stderr, returncode = get_command_data(cmd)
            
            requests = [
                get_req('GET', 'subprocess.Popen'),
                get_req('GET', 'subprocess.PIPE'),
                get_req('GET', 'subprocess.PIPE'),
                get_req('CALL', 'obj1', [(cmd,), {'shell': True, 'stderr': subprocess.PIPE, 'stdout': subprocess.PIPE}]),
                get_req('GET', 'obj2.communicate'),
                get_req('CALL', 'obj3', [tuple(), {}]),
                get_req('GET', 'obj4.__iter__'),
                get_req('CALL', 'obj5', [tuple(), {}]),
                get_req('GET', 'obj6.next' if six.PY2 else 'obj6.__next__'),
                get_req('CALL', 'obj7', [tuple(), {}]),
                get_req('CALL', 'obj7', [tuple(), {}]),
                get_req('CALL', 'obj7', [tuple(), {}]),
                get_req('GET', 'obj2.returncode'),
                b'',
            ]
            
            responses = [
                get_resp('200 OK', base.RemoteObjRef('obj1')),
                get_resp('200 OK', subprocess.PIPE),
                get_resp('200 OK', subprocess.PIPE),
                get_resp('200 OK', base.RemoteObjRef('obj2')),
                get_resp('200 OK', base.RemoteObjRef('obj3')),
                get_resp('200 OK', base.RemoteObjRef('obj4')),
                get_resp('200 OK', base.RemoteObjRef('obj5')),
                get_resp('200 OK', base.RemoteObjRef('obj6')),
                get_resp('200 OK', base.RemoteObjRef('obj7')),
                get_resp('200 OK', stdout),
                get_resp('200 OK', stderr),
                get_resp('400 Error', StopIteration()),
                get_resp('200 OK', returncode),
            ]
            
            clientsocket.recv.side_effect = iter(responses)
            
            socket.socket.return_value = clientsocket
            
            result = transport.run_cmd(cmd)
        
        self.assertEqual(clientsocket.connect.call_count, 1)
        self.assertEqual(clientsocket.connect.call_args_list[0][0][0], '/tmp/errand-boy')
        
        self.assertEqual(clientsocket.sendall.call_count, len(requests)-1)
        
        for i, request in enumerate(requests[:-1]):
            self.assertEqual(clientsocket.sendall.call_args_list[i][0][0], request)
        
        self.assertEqual(clientsocket.recv.call_count, len(responses))
        
        self.assertEqual(result[0], stdout)
        self.assertEqual(result[1], stderr)
        self.assertEqual(result[2], returncode)


class UNIXSocketTransportServerTestCase(BaseTestCase):
    def test(self):
        transport = unixsocket.UNIXSocketTransport()
        
        with self.socket_patcher as socket,\
                self.reduce_socket_patcher as reduce_socket,\
                self.rebuild_socket_patcher as rebuild_socket,\
                self.multiprocessing_patcher as multiprocessing,\
                self.subprocess_patcher as mock_subprocess,\
                self.uuid_patcher as uuid:
            mock_subprocess.PIPE = subprocess.PIPE
            
            serversocket = mock.Mock()
            
            
            cmd = 'ls -al'
            
            stdout, stderr, returncode = get_command_data(cmd)
            
            process = mock.Mock()
            process.communicate.return_value = stdout, stderr
            process.returncode = returncode
            
            mock_subprocess.Popen.return_value = process
            
            uuid.side_effect = ('obj%s' % i for i in six.moves.range(1, 20))
            
            requests = [
                get_req('GET', 'subprocess.Popen'),
                get_req('GET', 'subprocess.PIPE'),
                get_req('GET', 'subprocess.PIPE'),
                get_req('CALL', 'obj1', [(cmd,), {'shell': True, 'stdout': subprocess.PIPE, 'stderr': subprocess.PIPE}]),
                get_req('GET', 'obj2.communicate'),
                get_req('CALL', 'obj3', [tuple(), {}]),
                get_req('GET', 'obj4.__iter__'),
                get_req('CALL', 'obj5', [tuple(), {}]),
                get_req('GET', 'obj6.next' if six.PY2 else 'obj6.__next__'),
                get_req('CALL', 'obj7', [tuple(), {}]),
                get_req('CALL', 'obj7', [tuple(), {}]),
                get_req('CALL', 'obj7', [tuple(), {}]),
                get_req('GET', 'obj2.returncode'),
                b'',
            ]
            
            responses = [
                get_resp('200 OK', base.RemoteObjRef('obj1')),
                get_resp('200 OK', subprocess.PIPE),
                get_resp('200 OK', subprocess.PIPE),
                get_resp('200 OK', base.RemoteObjRef('obj2')),
                get_resp('200 OK', base.RemoteObjRef('obj3')),
                get_resp('200 OK', base.RemoteObjRef('obj4')),
                get_resp('200 OK', base.RemoteObjRef('obj5')),
                get_resp('200 OK', base.RemoteObjRef('obj6')),
                get_resp('200 OK', base.RemoteObjRef('obj7')),
                get_resp('200 OK', stdout),
                get_resp('200 OK', stderr),
                get_resp('400 Error', StopIteration()),
                get_resp('200 OK', returncode),
            ]
            
            clientsocket = mock.Mock()
            clientsocket.recv.side_effect = iter(requests)
            
            serversocket.accept.return_value = clientsocket, ''
            
            socket.socket.return_value = serversocket
            
            reduce_socket.return_value = (reduce_socket, ('I\'m a socket, NOT!', '', '', '',))
            rebuild_socket.return_value = clientsocket
            
            
            mock_Pool = mock.Mock()
            mock_Pool.apply_async.side_effect = lambda f, args=(), kwargs={}: f(*args, **kwargs)
            multiprocessing.Pool.return_value = mock_Pool
            
            transport.run_server(max_accepts=1)
        
        self.assertEqual(serversocket.bind.call_count, 1)
        self.assertEqual(serversocket.bind.call_args_list[0][0][0], '/tmp/errand-boy')
        
        self.assertEqual(serversocket.listen.call_count, 1)
        self.assertEqual(serversocket.listen.call_args_list[0][0][0], 5)
        
        self.assertEqual(reduce_socket.call_count, 1)
        self.assertEqual(reduce_socket.call_args_list[0][0][0], clientsocket)
        
        self.assertEqual(rebuild_socket.call_count, 1)
        self.assertEqual(rebuild_socket.call_args_list[0][0], reduce_socket.return_value[1])
        
        self.assertEqual(clientsocket.recv.call_count, len(requests))
        
        self.assertEqual(mock_subprocess.Popen.call_count, 1)
        self.assertEqual(mock_subprocess.Popen.call_args_list[0][0][0], cmd)
        
        self.assertEqual(clientsocket.sendall.call_count, len(responses))
        for i, response in enumerate(responses):
            self.assertEqual(clientsocket.sendall.call_args_list[i][0][0], response)
    
    def test_max_accepts_zero(self):
        transport = unixsocket.UNIXSocketTransport()
        
        with self.socket_patcher as socket,\
                self.reduce_socket_patcher as reduce_socket,\
                self.rebuild_socket_patcher as rebuild_socket,\
                self.multiprocessing_patcher as multiprocessing,\
                self.subprocess_patcher as mock_subprocess,\
                self.uuid_patcher as uuid:
            mock_subprocess.PIPE = subprocess.PIPE
            
            serversocket = mock.Mock()
            
            
            cmd = 'ls -al'
            
            stdout, stderr, returncode = get_command_data(cmd)
            
            process = mock.Mock()
            process.communicate.return_value = stdout, stderr
            process.returncode = returncode
            
            mock_subprocess.Popen.return_value = process
            
            uuid.side_effect = ('obj%s' % i for i in six.moves.range(1, 20))
            
            requests = [
                get_req('GET', 'subprocess.Popen'),
                get_req('GET', 'subprocess.PIPE'),
                get_req('GET', 'subprocess.PIPE'),
                get_req('CALL', 'obj1', [(cmd,), {'shell': True, 'stdout': subprocess.PIPE, 'stderr': subprocess.PIPE}]),
                get_req('GET', 'obj2.communicate'),
                get_req('CALL', 'obj3', [tuple(), {}]),
                get_req('GET', 'obj4.__iter__'),
                get_req('CALL', 'obj5', [tuple(), {}]),
                get_req('GET', 'obj6.next' if six.PY2 else 'obj6.__next__'),
                get_req('CALL', 'obj7', [tuple(), {}]),
                get_req('CALL', 'obj7', [tuple(), {}]),
                get_req('CALL', 'obj7', [tuple(), {}]),
                get_req('GET', 'obj2.returncode'),
                b'',
            ]
            
            responses = [
                get_resp('200 OK', base.RemoteObjRef('obj1')),
                get_resp('200 OK', subprocess.PIPE),
                get_resp('200 OK', subprocess.PIPE),
                get_resp('200 OK', base.RemoteObjRef('obj2')),
                get_resp('200 OK', base.RemoteObjRef('obj3')),
                get_resp('200 OK', base.RemoteObjRef('obj4')),
                get_resp('200 OK', base.RemoteObjRef('obj5')),
                get_resp('200 OK', base.RemoteObjRef('obj6')),
                get_resp('200 OK', base.RemoteObjRef('obj7')),
                get_resp('200 OK', stdout),
                get_resp('200 OK', stderr),
                get_resp('400 Error', StopIteration()),
                get_resp('200 OK', returncode),
            ]
            
            clientsocket = mock.Mock()
            clientsocket.recv.side_effect = iter(requests)
            
            serversocket.accept.side_effect = iter([(clientsocket, '',)])
            
            socket.socket.return_value = serversocket
            
            reduce_socket.return_value = (reduce_socket, ('I\'m a socket, NOT!', '', '', '',))
            rebuild_socket.return_value = clientsocket
            
            
            mock_Pool = mock.Mock()
            mock_Pool.apply_async.side_effect = lambda f, args=(), kwargs={}: f(*args, **kwargs)
            multiprocessing.Pool.return_value = mock_Pool
            
            # serve until an error happens (in ths case StopIteration from running
            # out of items in clientsocket.recv.side_effect)
            try:
                transport.run_server(max_accepts=0)
            except StopIteration:
                pass
        
        self.assertEqual(serversocket.bind.call_count, 1)
        self.assertEqual(serversocket.bind.call_args_list[0][0][0], '/tmp/errand-boy')
        
        self.assertEqual(serversocket.listen.call_count, 1)
        self.assertEqual(serversocket.listen.call_args_list[0][0][0], 5)
        
        self.assertEqual(reduce_socket.call_count, 1)
        self.assertEqual(reduce_socket.call_args_list[0][0][0], clientsocket)
        
        self.assertEqual(rebuild_socket.call_count, 1)
        self.assertEqual(rebuild_socket.call_args_list[0][0], reduce_socket.return_value[1])
        
        self.assertEqual(clientsocket.recv.call_count, len(requests))
        
        self.assertEqual(mock_subprocess.Popen.call_count, 1)
        self.assertEqual(mock_subprocess.Popen.call_args_list[0][0][0], cmd)
        
        self.assertEqual(clientsocket.sendall.call_count, len(responses))
        for i, response in enumerate(responses):
            self.assertEqual(clientsocket.sendall.call_args_list[i][0][0], response)



import collections
import pickle

from errand_boy import constants
from errand_boy.transports import base


commands = {
    'cat /dev/null': ([("", "",)], 0),
    'ls -al': ([("""total 96
drwxrwxr-x  5 me me 4096 Aug  5 10:50 .
drwxrwxr-x 11 me me 4096 Mar 28 15:40 ..
drwxrwxr-x  3 me me 4096 Jan 27  2014 errand_boy
drwxrwxr-x  8 me me 4096 Aug  4 19:45 .git
-rw-rw-r--  1 me me  315 Jan 20  2014 .gitignore
-rw-rw-r--  1 me me 1482 Jan 20  2014 LICENSE
-rw-rw-r--  1 me me 1972 Jan 27  2014 README.rst
-rwxrwxr-x  1 me me 1171 Mar  3 02:29 setup.py
drwxrwxr-x  2 me me 4096 Aug  5 10:48 tests
-rw-rw-r--  1 me me  166 Jan 27  2014 .travis.yml""", "",)], 0),
}


def get_command_data(cmd):
    result = commands[cmd]
    
    stdout, stderr = zip(*result[0])
    
    return ''.join(stdout), ''.join(stderr), result[1]

def get_req(method, path, obj=None):
    if obj is not None:
        try:
            obj[1] = collections.OrderedDict(sorted(obj[1].items(), key=lambda t: t[0]))
        except:
            pass
        
        obj = pickle.dumps(obj)
        
        return ("%s %s\r\nContent-Length: %s\r\n\r\n" % (method, path, len(obj))).encode('utf-8') + obj
    else:
        return ("%s %s\r\nContent-Length: 0\r\n" % (method, path,)).encode('utf-8')

def get_resp(status, obj=None):
    if obj is not None:
        try:
            obj[1] = collections.OrderedDict(sorted(obj[1].items(), key=lambda t: t[0]))
        except:
            pass
        
        obj = pickle.dumps(obj)
        
        return ("%s\r\nContent-Length: %s\r\n\r\n" % (status, len(obj))).encode('utf-8') + obj
    else:
        return ("%s\r\nContent-Length: 0\r\n" % (status,)).encode('utf-8')




try:
    from unittest import mock
except ImportError:
    import mock
import unittest

from errand_boy import run
from errand_boy.transports import base, mock as mock_transport, unixsocket


class BaseTestCase(unittest.TestCase):
    subprocess_patcher = mock.patch.object(base, 'subprocess', autospec=True)
    multiprocessing_patcher = mock.patch.object(base, 'multiprocessing', autospec=True)
    sys_patcher = mock.patch.object(run, 'sys', autospec=True)
    uuid_patcher = mock.patch.object(base.uuid, 'uuid4', autospec=True)
    
    UNIXSocketTransport_patcher = mock.patch.object(unixsocket, 'UNIXSocketTransport', autospec=True)
    socket_patcher = mock.patch.object(unixsocket, 'socket', autospec=True)
    reduce_socket_patcher = mock.patch.object(unixsocket, 'reduce_socket', autospec=True)
    rebuild_socket_patcher = mock.patch.object(unixsocket, 'rebuild_socket', autospec=True)



import collections
import logging
import multiprocessing
import numbers
import pickle
import signal
import subprocess
import six
import sys
import uuid

from .. import constants
from .. import __version__


logger = logging.getLogger(__name__)

try:
    from setproctitle import setproctitle
except ImportError:
    logger.info('Cannot set process name.')
    setproctitle = lambda title: None


class DisconnectedError(Exception):
    pass


class ClientSession(object):
    def __init__(self, transport):
        self.transport = transport
        self.connection = transport.client_get_connection()
    
    def __getattr__(self, name):
        return RemoteObjWrapper(self, name)
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.transport.client_close(self.connection)
        except Exception as e:
            logger.exception(e)
        
        # we don't want to hide AttributeErrors, etc. from end user
        return False


class RemoteObjRef(object):
    def __init__(self, name):
        if hasattr(name, 'decode'):
            name = name.decode('utf-8')
        self.name = name
    
    def __str__(self):
        return self.name


class RemoteObjWrapper(object):
    _patch_functions = ['next', '__next__', '__iter__']
    
    def __init__(self, session, name):
        self.session = session
        self.name = name
        
        self._property_cache = {}
        
        for name in self._patch_functions:
            self._property_cache[name] = None
    
    def __getattr__(self, name):
        session = self.session
        ret = session.transport.send_get_request(session.connection, self.name, name)
        if isinstance(ret, RemoteObjRef):
            ret = RemoteObjWrapper(session, ret.name)
        return ret
    
    def __call__(self, *args, **kwargs):
        session = self.session
        ret = session.transport.send_call_request(session.connection, self.name, *args, **kwargs)
        if isinstance(ret, RemoteObjRef):
            ret = RemoteObjWrapper(session, ret.name)
        return ret
    
    @property
    def next(self):
        name = 'next'
        val = self._property_cache[name]
        if val:
            return val
        self._property_cache[name] = self.__getattr__(name)
        return self._property_cache[name]
    
    @property
    def __next__(self):
        name = '__next__'
        val = self._property_cache[name]
        if val:
            return val
        self._property_cache[name] = self.__getattr__(name)
        return self._property_cache[name]
    
    @property
    def __iter__(self):
        name = '__iter__'
        val = self._property_cache[name]
        if val:
            return val
        self._property_cache[name] = self.__getattr__(name)
        return self._property_cache[name]


class Request(object):
    def __init__(self, method, path, headers, body):
        self.method = method
        self.path = path
        self.headers = headers
        self.body = body


class Response(object):
    def __init__(self, status, headers, body):
        self.status = status
        self.headers = headers
        self.body = body


def worker_init(*args):
    name = multiprocessing.current_process().name
    logger.debug('Worker initialized: {}'.format(name))
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    setproctitle('errand-boy worker process {}'.format(name.split('-')[1]))


def worker(self, connection):
    logger.debug('worker connected')
    self.server_handle_client(connection)


class BaseTransport(object):
    """
    Base class providing functionality common to all transports.
    """
    
    def __init__(self):
        pass
    
    def connection_to_string(self, connection):
        return repr(connection)
    
    def server_get_connection(self):
        pass
    
    def server_recv(self, connection):
        pass
    
    def server_send(self, connection, data):
        pass
    
    def server_close(self, connection):
        pass
    
    def translate_obj(self, exposed_locals, val):
        if isinstance(val, RemoteObjRef):
            val = exposed_locals[val.name]
        return val
    
    def server_serialize(self, exposed_locals, obj):
        if obj is not None and not isinstance(obj, six.string_types+(numbers.Number, BaseException)):
            name = six.text_type(uuid.uuid4())
            exposed_locals[name] = obj
            obj = RemoteObjRef(name)
        return obj
    
    def server_handle_client(self, connection):
        connection = self.server_deserialize_connection(connection)
        
        logger.debug('server_handle_client: {}'.format(self.connection_to_string(connection)))
        
        exposed_locals = {'subprocess': subprocess}
        
        while True:
            # need to close connection when client not listening
            try:
                 request = self.get_request(connection)
            except DisconnectedError:
                 break
            
            raised = False
            obj = None
            
            if request.method == 'GET':
                name, attr = request.path.split('.')
                try:
                    obj = getattr(exposed_locals[name], attr)
                except KeyError as e:
                    obj = e
                    raised = True
            elif request.method == 'CALL':
                name = request.path
                obj = exposed_locals[name]
                args, kwargs = pickle.loads(request.body)
                
                args = [self.translate_obj(exposed_locals, arg) for arg in args]
                
                for key in kwargs:
                    kwargs[key] = self.translate_obj(exposed_locals, kwargs[key])
                
                try:
                    obj = obj(*args, **kwargs)
                except Exception as e:
                    obj = e
                    raised = True
            
            obj = self.server_serialize(exposed_locals, obj)
            
            self.send_response(connection, obj, raised=raised)
        
        self.server_close(connection)
    
    def server_accept(self, serverconnection):
        pass
    
    def server_deserialize_connection(self, connection):
        return connection
    
    def server_serialize_connection(self, connection):
        return connection
    
    def run_server(self, pool_size=10, max_accepts=5000, max_child_tasks=100):
        setproctitle('errand-boy master process')
        
        serverconnection = self.server_get_connection()
        
        logger.info('Accepting connections: {}'.format(self.connection_to_string(serverconnection)))
        logger.info('pool_size: {}'.format(pool_size))
        logger.info('max_accepts: {}'.format(max_accepts))
        logger.info('max_child_tasks: {}'.format(max_child_tasks))
        
        pool = multiprocessing.Pool(pool_size, worker_init, tuple(), max_child_tasks)
        
        connections = []
        
        remaining_accepts = max_accepts
        
        if not remaining_accepts:
            remaining_accepts = True
        
        try:
            while remaining_accepts:
                connection = self.server_accept(serverconnection)
                
                logger.info('Accepted connection from: {}'.format(self.connection_to_string(connection)))
                
                result = pool.apply_async(worker, [self, self.server_serialize_connection(connection)])
                
                connection = None
                
                if remaining_accepts is not True:
                    remaining_accepts -= 1
        except KeyboardInterrupt:
            logger.info('Received KeyboardInterrupt')
            pool.terminate()
        except Exception as e:
            logger.exception(e)
            pool.terminate()
            raise
        finally:
            pool.close()
            pool.join()
    
    def client_get_connection(self):
        pass
    
    def client_send(self, connection, command_string):
        pass
    
    def client_recv(self, connection):
        pass
    
    def client_close(self, connection):
        pass
    
    def send_algo(self, connection, send_func, first_line, headers=None, body=None):
        CRLF = constants.CRLF
        msg = [first_line]
        msg.append(CRLF)
        
        if headers:
            for name, val in headers:
                msg.append('{}: {}'.format(name, val))
                msg.append(CRLF)
        
        msg.append('Content-Length: {}'.format(len(body)))
        msg.append(CRLF)
        
        if body:
            msg.append(CRLF)
            msg.append(body)
        
        msg = [s.encode('utf-8') if hasattr(s, 'encode') else s for s in msg]
        
        msg = b''.join(msg)
        
        return send_func(connection, msg)
    
    def send_request(self, connection, method, path, body=''):
        first_line = "{method} {path}".format(method=method, path=path)
        self.send_algo(connection, self.client_send, first_line, headers=None, body=body)
        
        resp = self.get_response(connection)
        
        obj = pickle.loads(resp.body)
        
        if resp.status == 400:
            raise obj
        
        return obj
    
    def send_get_request(self, connection, prefix, name):
        return self.send_request(connection, 'GET', prefix+'.'+name)
    
    def send_call_request(self, connection, name, *args, **kwargs):
        kwargs = collections.OrderedDict(sorted(kwargs.items(), key=lambda t: t[0]))
        body = pickle.dumps([args, kwargs])
        
        return self.send_request(connection, 'CALL', name, body=body)
    
    def recv_algo(self, connection, recv_func):
        CRLF = constants.CRLF
        
        lines = []
        data = b''
        
        content_length = 0
        
        while True:
            new_data = recv_func(connection, 4096)
            
            if not new_data:
                raise DisconnectedError()
            
            data += new_data
            
            if CRLF in data:
                split_data = data.split(CRLF)
                lines.extend(split_data[:-1])
                data = split_data[-1]
            
            if lines and content_length is None:
                for line in lines:
                    try:
                        header, val = line.split(b': ')
                        if header.lower() == 'content-length':
                            content_length = int(val)
                            break
                    except ValueError:
                        pass
            
            if content_length == 0 and len(lines) > 1:
                break
            
            if len(lines) > 3 and lines[-1] == '' and lines[-2] == '':
                # needs length of body, minus already fetched data
                data += clientsocket.recv(connection, (2+content_length)-len(data))
                break
        
        if data:
            data = CRLF + data
        data = CRLF.join(lines) + data
        
        try:
            headers, body = data.split(constants.CRLF+constants.CRLF)
        except ValueError:
            headers = data
            body = b''
        
        headers = headers.split(constants.CRLF)
        headers = [header.decode('utf-8') for header in headers]
        first_line = headers[0]
        headers = headers[1:]
        
        headers = [header.split(': ') for header in headers]
        return first_line, headers, body
    
    def get_request(self, connection):
        first_line, headers, body = self.recv_algo(connection, self.server_recv)
        
        method, path = first_line.split(' ', 1)
        
        return Request(method, path, headers, body)
    
    def send_response(self, connection, obj, raised=False):
        body = pickle.dumps(obj)
        
        first_line = '200 OK' if not raised else '400 Error'
        
        return self.send_algo(connection, self.server_send, first_line, body=body)
    
    def get_response(self, connection):
        first_line, headers, body = self.recv_algo(connection, self.client_recv)
        
        status = int(first_line.split()[0])
        
        return Response(status, headers, body)
    
    def get_session(self):
        return ClientSession(self)
    
    def run_cmd(self, command_string):
        with self.get_session() as session:
            subprocess = session.subprocess
            
            process = subprocess.Popen(command_string, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            
            stdout, stderr = process.communicate()
            
            returncode = process.returncode
        
        return stdout, stderr, returncode



import errand_boy
from errand_boy import run
from errand_boy.transports import unixsocket

from .base import mock, BaseTestCase


class MainTestCase(BaseTestCase):
    def test_client(self):
        argv = ['/srv/errand-boy/errand_boy/run.py', 'ls', '-al']
        cmd = ' '.join(argv[1:])
        
        with self.UNIXSocketTransport_patcher as UNIXSocketTransport,\
                self.sys_patcher as mocked_sys:
            mock_process = mock.Mock()
            mock_process.returncode = 0
            
            stdout = 'foo'
            stderr = 'bar'
            
            transport = mock.Mock()
            transport.run_cmd.return_value = stdout, stderr, mock_process.returncode
            
            UNIXSocketTransport.return_value = transport
            
            run.main(argv)
        
        self.assertEqual(transport.run_cmd.call_count, 1)
        self.assertEqual(transport.run_cmd.call_args_list[0][0][0], cmd)
        
        self.assertEqual(mocked_sys.stdout.write.call_count, 1)
        self.assertEqual(mocked_sys.stdout.write.call_args_list[0][0][0], stdout)
        
        self.assertEqual(mocked_sys.stderr.write.call_count, 1)
        self.assertEqual(mocked_sys.stderr.write.call_args_list[0][0][0], stderr)
        
        self.assertEqual(mocked_sys.exit.call_count, 1)
        self.assertEqual(mocked_sys.exit.call_args_list[0][0][0], mock_process.returncode)
    
    def test_server_no_options(self):
        argv = ['/srv/errand-boy/errand_boy/run.py']
        
        with self.UNIXSocketTransport_patcher as UNIXSocketTransport:
            transport = mock.Mock()
            
            UNIXSocketTransport.return_value = transport
            
            run.main(argv)
        
        self.assertEqual(transport.run_server.call_count, 1)
        self.assertEqual(transport.run_server.call_args_list[0][0], tuple())
        self.assertEqual(transport.run_server.call_args_list[0][1], {'max_accepts': 5000, 'max_child_tasks': 100, 'pool_size': 10})
    
    def test_server_with_options(self):
        argv = ['/srv/errand-boy/errand_boy/run.py', '--max-accepts', '5']
        
        with self.UNIXSocketTransport_patcher as UNIXSocketTransport:
            transport = mock.Mock()
            
            UNIXSocketTransport.return_value = transport
            
            run.main(argv)
        
        self.assertEqual(transport.run_server.call_count, 1)
        self.assertEqual(transport.run_server.call_args_list[0][0], tuple())
        self.assertEqual(transport.run_server.call_args_list[0][1], {'max_accepts': int(argv[2]), 'max_child_tasks': 100, 'pool_size': 10})




import argparse
import importlib
import logging
import logging.config
import sys

from . import __version__


parser = argparse.ArgumentParser(description='Start errand-boy.')
parser.add_argument('-t', '--transport', dest='transport', nargs='?',
           default='errand_boy.transports.unixsocket.UNIXSocketTransport',
           help='Python path to the transport to use.')
parser.add_argument('--pool-size', dest='pool_size', nargs='?', type=int,
           default=10,
           help='Number of worker processes to use.')
parser.add_argument('--max-accepts', dest='max_accepts', nargs='?', type=int,
           default=5000,
           help='Max number of connections the server will accept.')
parser.add_argument('--max-child-tasks', dest='max_child_tasks', nargs='?', type=int,
           default=100,
           help='Max number of tasks each child process will handle before being replaced.')
parser.add_argument('command', nargs=argparse.REMAINDER)
parser.add_argument('--version', action='version', version=__version__)

class MaxLevelFilter(logging.Filter):
    def __init__(self, level):
        self._level = level

    def filter(self, rec):
        return rec.levelno <= self._level

LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'filters': {
        'stdout': {
            '()': MaxLevelFilter,
            'level': 'WARNING',
        },
    },
    'formatters': {
        'verbose': {
            'format': '%(levelname)s %(asctime)s %(process)d %(thread)d %(name)s:%(lineno)s %(funcName)s() %(message)s'
        },
    },
    'handlers': {
        'stdout': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'stream': sys.stdout,
            'formatter': 'verbose',
        },
        'stderr': {
            'level': 'ERROR',
            'class': 'logging.StreamHandler',
            'stream': sys.stderr,
            'formatter': 'verbose',
        },
    },
    'loggers': {
        'errand_boy': {
            'handlers': ['stderr', 'stdout'],
            'level': 'INFO',
            'propagate': False,
        },
    },
}

def main(argv):
    parsed_args = parser.parse_args(argv[1:])
    
    mod, klass = parsed_args.transport.rsplit('.', 1)
    
    transport = getattr(importlib.import_module(mod), klass)()
    
    command = parsed_args.command
    
    if not command:
        logging.config.dictConfig(LOGGING)
        
        transport.run_server(
            pool_size=parsed_args.pool_size,
            max_accepts=parsed_args.max_accepts,
            max_child_tasks=parsed_args.max_child_tasks
        )
    else:
        stdout, stderr, returncode = transport.run_cmd(' '.join(command))
        
        sys.stdout.write(stdout)
        sys.stderr.write(stderr)
        
        sys.exit(returncode)

if __name__ == '__main__':
    main(sys.argv)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import errand_boy

package_name = 'errand_boy'
test_package_name = 'tests'

setup(
    name='errand-boy',
    version=errand_boy.__version__,
    description="Establish a connection to the errand-boy deamon to execute commands without the memory overhead incurred by os.fork().",
    author='Seán Hayes',
    author_email='sean@seanhayes.name',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    keywords='os fork process memory celery',
    url='https://github.com/SeanHayes/errand-boy',
    download_url='https://github.com/SeanHayes/errand-boy',
    license='BSD',
    packages=find_packages(),
    install_requires=['eventlet', 'six'],
    tests_require=[
        'mock'
    ],
    include_package_data=True,
    test_suite = test_package_name,
)




import subprocess

import errand_boy
from errand_boy.transports import base, mock as mock_transport

from .base import mock, BaseTestCase
from .data import get_command_data

class MockTransportTestCase(BaseTestCase):
    def test(self):
        transport = mock_transport.MockTransport()
        
        with self.multiprocessing_patcher as multiprocessing,\
                self.subprocess_patcher as mock_subprocess:
            mock_subprocess.PIPE = subprocess.PIPE
            
            cmd = 'ls -al'
            
            stdout, stderr, returncode = get_command_data(cmd)
            
            process = mock.Mock()
            process.communicate.return_value = stdout, stderr
            process.returncode = returncode
            
            mock_subprocess.Popen.return_value = process
            
            mock_Pool = mock.Mock()
            mock_Pool.apply_async.side_effect = lambda f, args=(), kwargs={}: f(*args, **kwargs)
            multiprocessing.Pool.return_value = mock_Pool
            
            result = transport.run_cmd(cmd)
        
        self.assertEqual(result[0], stdout)
        self.assertEqual(result[1], stderr)
        self.assertEqual(result[2], returncode)
        
        self.assertEqual(mock_subprocess.Popen.call_count, 1)
        self.assertEqual(mock_subprocess.Popen.call_args_list[0][0][0], cmd)
